#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
@Modifider: Cui Jinghao
@contact: cuijinghao@tgqs.net
@file: QuantumCircuit.py
@time: 2024/06/28 14:40

reconstructed quantum circuit file
"""
import tgqSim
from tgqSim.circuit.common_gates import CommonGate
import tgqSim.draw_circuit_tools as tools
from tgqSim.circuit.common_gates import (BASE_SINGLE_GATE,
                                               BASE_DOUBLE_GATE,
                                               BASE_TRIPLE_GATE,
                                               MEASURE, )
from tgqSim.device.noise_models import NOISE_MAPPER
from tgqSim.utils.visualization import to_text_diag
from tgqSim.utils.logger import Logger
import logging
from typing import Union
import random, os, ctypes
import numpy as np
import matplotlib.pyplot as plt
# import GPUtil, datetime
# from numba import njit, prange

log = Logger(name="tgqsim", level=logging.DEBUG, log_file="app.log")


class Float2(ctypes.Structure):
    _fields_ = [("x", ctypes.c_float),
                ("y", ctypes.c_float)]

class Double2(ctypes.Structure):
    _fields_ = [("x", ctypes.c_double),
                ("y", ctypes.c_double)]

class GateInfo(ctypes.Structure):
    _fields_ = [
        ("gateName", ctypes.c_char_p),
        ("actionPos", ctypes.POINTER(ctypes.c_int)),
        ("theta", ctypes.c_double)
    ]

class Qubit:
    def __init__(self, name=None, idx=None, pos=None):
        self.label = 'p' if not name else name
        self.name = f'p_{idx+100}' if not name else f"{name}_{idx}"
        self.pos = pos + idx
        self.idx = idx

    # def __getattr__(self, item):
    #     if item == self.name:
    #         return self.idx
    #     if item == self.label+"["+str(self.pos)+"]":
    #         return self.idx

    def __str__(self):
        return str(self.name)


class QuantumCircuit:
    def __init__(self, qbit_number=0):
        self.qbit_list = []
        self.cbit_list = []
        self.gate_list = []
        self.qbit_register = {}
        self.width = qbit_number
        self.state = []
        self.circuit_diag = []
        # self.isgpu = False
        # self.isnpu = False
        self.noise_circuit = []
        self.current_gate_info = ()
        self.diagram_str = []
        # contains gates's display_names
        self.displayname_list = []

    # todo: implement a visualization method for circuit, use displayname_list to denote
    def __str__(self):
        return to_text_diag(self.displayname_list, self.width)

    def add_single_gate(self, gate_pos: Union[int, str], display_name: tuple, *gate_info):
        """
        添加单比特门序列
        :param gate_pos:
        :param gate_info: 门类型，门参数
        :param display_name:
        :return:
        """
        tmp_pos = gate_pos
        if isinstance(gate_pos, list):
            tmp_pos = gate_pos[0]
        if tmp_pos < 0 or tmp_pos >= self.width:
            raise ValueError("添加的位置不能为负数或不能超过线路设置的比特数")
        if len(gate_info) >= 1:
            if gate_info[0] not in BASE_SINGLE_GATE:
                raise ValueError("添加的门名称在框架支持中，请检查添加门的名称")
        if isinstance(gate_pos, int):
            gatename = gate_info[0].upper()
            self.circuit_diag.append([(gatename, gate_pos)])
        else:
            gatename = gate_info[0].upper()
            self.circuit_diag.append([(gatename, int(gate_pos))])

        self.current_gate_info = (gate_pos, gate_info)
        self.noise_circuit.append((gate_pos, gate_info))
        self.gate_list.append((gate_pos, gate_info))
        self.displayname_list.append((gate_pos, display_name, gate_info))
        # print("over")
        return self
        # todo 通过比特名称访问
        # if isinstance(gate_pos, str):
        #     return self.gate_list.append((gate_pos, gate_type))

    def add_double_gate(self, gate_pos_0:int, gate_pos_1: int, display_name: str, *gate_info):
        """
        添加两比特门序列
        :param gate_pos_0: 控制位
        :param gate_pos_1: 目标位
        :param gate_info: 门类型，门参数
        :param display_name:
        :return:
        """
        min_pos, max_pos = min([gate_pos_0, gate_pos_1]), max([gate_pos_0, gate_pos_1])
        if min_pos < 0 or max_pos >= self.width:
            raise ValueError("添加的位置不能为负数或不能超过线路设置的比特数")
        if len(gate_info) >= 1:
            if gate_info[0] not in BASE_DOUBLE_GATE:
                raise ValueError("添加的门名称在框架支持中，请检查添加门的名称")
        if gate_info[0] in ["rxx", 'ryy', 'rzz', 'iswap', 'syc']:
            self.circuit_diag.append([(gate_info[0].upper(), (gate_pos_1, gate_pos_0))])
        else:
            self.circuit_diag.append([(gate_info[0].upper(), gate_pos_1, gate_pos_0)])
        self.current_gate_info = ([gate_pos_0, gate_pos_1], gate_info)
        self.noise_circuit.append(([gate_pos_0, gate_pos_1], gate_info))
        self.gate_list.append(([gate_pos_0, gate_pos_1], gate_info))
        self.displayname_list.append(([gate_pos_0, gate_pos_1], display_name, gate_info))
        return self

    def add_triple_gate(self, gate_pos: list, display_name: str, *gate_info):
        """
        添加三比特门序列
        :param gate_pos:
        :param gate_info:
        :param display_name:
        :return:
        """
        min_pos, max_pos = min(gate_pos), max(gate_pos)
        if min_pos < 0 or max_pos >= self.width:
            raise ValueError("添加的位置不能为负数或不能超过线路设置的比特数")
        if len(gate_info) >= 1:
            if gate_info[0] not in BASE_TRIPLE_GATE:
                raise ValueError("添加的门名称在框架支持中，请检查添加门的名称")
        self.circuit_diag.append([(gate_info[0].upper(), gate_pos[2], gate_pos[0], gate_pos[1])])
        self.current_gate_info = (gate_pos, gate_info)
        self.noise_circuit.append((gate_pos, gate_info))
        self.gate_list.append((gate_pos, gate_info))
        self.displayname_list.append((gate_pos, display_name, gate_info))
        return self

    # todo: define add measure gate method
    def add_measure_gate(self, gate_pos: Union[list, int], display_name: str, *gate_info):
        if isinstance(gate_pos, list):
            min_pos, max_pos = min(gate_pos), max(gate_pos)
        else:
            min_pos, max_pos = gate_pos, gate_pos
        if min_pos < 0 or max_pos >= self.width:
            raise ValueError("添加的位置不能为负数或不能超过线路设置的比特数")
        if len(gate_info) >= 1:
            if gate_info[0] not in MEASURE:
                raise ValueError("添加的门名称在框架支持中，请检查添加门的名称")
        self.gate_list.append((gate_pos, gate_info))
        self.displayname_list.append((gate_pos, display_name, gate_info))
        return self

    # append single gate
    def append(self, gate: CommonGate):
        """
        append gate to circuit
        Returns:
        three options: single,double and triple

        """
        # print(gate)

        # if gate['name'] in MEASURE:
        if gate.name in MEASURE:

            # todo:
            # self.add_measure_gate(gate['qbit'], gate['name'])
            self.add_measure_gate(gate.qbit, gate.display_name, gate.name)

        # elif gate['name'] in BASE_SINGLE_GATE:
        elif gate.name in BASE_SINGLE_GATE:

            # if gate['theta'] is None:
            if gate.theta is None:

                self.add_single_gate(gate.qbit, gate.display_name, gate.name)

            else:
                if isinstance(gate.theta, tuple):
                    self.add_single_gate(gate.qbit,
                                         gate.display_name,
                                         gate.name,
                                         *gate.theta)
                else:
                    # print('gate theta before append is {}'.format(gate['theta']))
                    self.add_single_gate(gate.qbit, gate.display_name ,gate.name, gate.theta)

        elif gate.name in BASE_DOUBLE_GATE:
            if not gate.theta:
                self.add_double_gate(gate.qbit[0], gate.qbit[1], gate.display_name, gate.name)
            else:
                self.add_double_gate(gate.qbit[0], gate.qbit[1], gate.display_name,
                                     gate.name, gate.theta)

        elif gate.name in BASE_TRIPLE_GATE:
            if not gate.theta:
                self.add_triple_gate(gate.qbit, gate.display_name, gate.name)
        # if gate.noise is not None:
        #     pass
        #     # noise looks like this [(qbit, (noise_type, error_rate))
        #     self.with_noise(gate.noise[0], gate.noise[1][0], gate.noise[1,1])

    def append_circuit(self, circuit: "QuantumCircuit"):
        self.gate_list += circuit.gate_list
        self.displayname_list += circuit.displayname_list

    def add_qubits(self, number: int, name: str = 'p'):
        """
        添加量子比特
        :param number: 比特数量
        :param name: 比特名称
        :return:
        """
        self.width += number
        if name not in self.qbit_register:
            self.qbit_register.update({name: number})
            sub_list = [Qubit(name=name, idx=i, pos=i+self.width) for i in range(number)]
        else:
            former_number = self.qbit_register[name]
            self.qbit_register.update({name: number+former_number})
            sub_list = [Qubit(name=name, idx=i+former_number, pos=i+self.width) for i in range(number)]
        self.qbit_list.append(sub_list)

    # moved setdevice method to simulator class

    # todo: add noise to each gate instead of to the whole circuit
    def with_noise(self, qbit:Union[list, int], noise_type:str, error_rate:Union[float, list]):
        """
        将当前的所有控制门都加上noise_type类型的噪声

        Args:
            noise_type (str): 噪声类型: "bit_flip","depolarize", "phase_flip", "phase_damp", "amplitude_damp"
            error_rate (Union[float, list]): 错误率
        """
        # self.noise_circuit += [self.current_gate_info]
        # gate_pos = self.current_gate_info[0]
        gate_pos = qbit
        if isinstance(gate_pos, int):
            self.noise_circuit += [(gate_pos, (noise_type, error_rate))]
            # self.circuit_diag.append([(NOISE_MAPPER[noise_type], gate_pos)])
        else:
            for gate in gate_pos:
                self.noise_circuit += [(gate, (noise_type, error_rate))]
                # self.circuit_diag.append([(NOISE_MAPPER[noise_type], gate)])

    def random_circuit(self, num_qubits, num_gates: Union[int, list]):
        """
        生成随机线路
        :param num_qubits:
        :param num_gates: 门数量，列表表示单比特，多比特门数量
        :return:
        """
        self.width = num_qubits
        if isinstance(num_gates, int):
            base_gate_list = BASE_SINGLE_GATE + BASE_DOUBLE_GATE + BASE_TRIPLE_GATE
            for _ in range(num_gates):
                gate_type = np.random.choice(base_gate_list)
                angle = np.random.uniform(-2*np.pi, 2*np.pi, size=3)
                if gate_type in BASE_SINGLE_GATE:
                    gate_pos = np.random.choice(np.array(range(num_qubits)), size=1)[0]
                elif gate_type in BASE_DOUBLE_GATE:
                    gate_pos = random.sample(range(num_qubits), 2)
                    gate_pos = list(gate_pos)
                elif gate_type in BASE_TRIPLE_GATE:
                    gate_pos = random.sample(range(num_qubits), 3)
                    gate_pos = list(gate_pos)
                else:
                    gate_pos = None
                self.gate_list.append([gate_pos, tuple([gate_type]) + tuple(angle)])
        elif isinstance(num_gates, list):
            single_gate_num = num_gates[0]
            multiple_gate_num = num_gates[1]
            for _ in range(single_gate_num):
                gate_type = np.random.choice(BASE_DOUBLE_GATE)
                angle = np.random.uniform(-2*np.pi, 2*np.pi, size=3)
                gate_pos = np.random.choice(np.array(range(num_qubits)), size=1)[0]
                self.gate_list.append([gate_pos, tuple([gate_type]) + tuple(angle)])
            for _ in range(multiple_gate_num):
                base_gate_list = BASE_DOUBLE_GATE+BASE_TRIPLE_GATE
                gate_type = np.random.choice(base_gate_list)
                angle = np.random.uniform(-2 * np.pi, 2 * np.pi, size=3)
                if gate_type in BASE_DOUBLE_GATE:
                    gate_pos = random.sample(range(num_qubits), 2)
                    gate_pos = list(gate_pos)
                elif gate_type in BASE_TRIPLE_GATE:
                    gate_pos = random.sample(range(num_qubits), 3)
                    gate_pos = list(gate_pos)
                else:
                    gate_pos = None
                self.gate_list.append([gate_pos, tuple([gate_type]) + tuple(angle)])
                random.shuffle(self.gate_list)

    # def show_quantum_circuit(self, plot_labels=True, **kwargs):
    #     """Use Matplotlib to plot a quantum circuit.
    #     kwargs    Can override plot_parameters
    #     """
    #     labels = []
    #     inits = {}
    #     for i in range(self.width):
    #         labels.append(f"q_{i}")
    #         inits[f"q_{i}"] = i
    #     plot_params = dict(scale = 1.0, fontsize = 14.5, linewidth = 2.0,
    #                         control_radius = 0.05, not_radius = 0.15,
    #                         swap_delta = 0.08, label_buffer = 0.8,
    #                         rectangle_delta = 0.3, box_pad=0.2)
    #     plot_params.update(kwargs)
    #     scale = plot_params['scale']
    #
    #     # Create labels from gates. This will become slow if there are a lot
    #     #  of gates, in which case move to an ordered dictionary
    #     if not labels:
    #         labels = []
    #         for i,gate in tools.enumerate_gates(self.circuit_diag, schedule=True):
    #             for label in gate[1:]:
    #                 if label not in labels:
    #                     labels.append(label)
    #
    #     nq = len(labels)
    #     nt = len(self.circuit_diag)
    #     wire_grid = np.arange(0.0, nq*scale, scale, dtype=float)
    #     gate_grid = np.arange(0.0, nt*scale, scale, dtype=float)
    #     gate_grid_index = [0.0 for _ in range(nq)]
    #     # print(gate_grid_index)
    #
    #     fig,ax = tools.setup_figure(nq,nt,gate_grid,wire_grid,plot_params)
    #
    #     measured = tools.measured_wires(self.circuit_diag, labels, schedule=True)
    #     tools.draw_wires(ax, nq, gate_grid, wire_grid, plot_params, 'k', measured)
    #
    #     if plot_labels:
    #         tools.draw_labels(ax, labels, inits, gate_grid, wire_grid, plot_params, 'k')
    #
    #     tools.draw_gates(ax, self.circuit_diag, labels, gate_grid_index, wire_grid, plot_params, measured, schedule=True)
    #     plt.show()
        # return ax

    def show_quantum_circuit(self, plot_labels=True, **kwargs):
        """Use Matplotlib to plot a quantum circuit.
        kwargs    Can override plot_parameters
        """
        labels = []
        inits = {}
        for i in range(self.width):
            labels.append(f"q_{i}")
            inits[f"q_{i}"] = i
        plot_params = dict(scale=1.0, fontsize=14.5, linewidth=2.0,
                           control_radius=0.05, not_radius=0.15,
                           swap_delta=0.08, label_buffer=0.8,
                           rectangle_delta=0.3, box_pad=0.2)
        plot_params.update(kwargs)
        scale = plot_params['scale']

        # Create labels from gates. This will become slow if there are a lot
        #  of gates, in which case move to an ordered dictionary
        if not labels:
            labels = []
            for i, gate in tools.enumerate_gates(self.circuit_diag, schedule=True):
                for label in gate[1:]:
                    if label not in labels:
                        labels.append(label)

        nq = len(labels)
        nt = len(self.circuit_diag)
        wire_grid = np.arange(0.0, nq * scale, scale, dtype=float)
        gate_grid = np.arange(0.0, nt * scale, scale, dtype=float)
        gate_grid_index = [0.0 for _ in range(nq)]
        # print(gate_grid_index)

        fig, ax = tools.setup_figure(nq, nt, gate_grid, wire_grid, plot_params)

        measured = tools.measured_wires(self.circuit_diag, labels, schedule=True)
        tools.draw_wires(ax, nq, gate_grid, wire_grid, plot_params, 'k', measured)

        if plot_labels:
            tools.draw_labels(ax, labels, inits, gate_grid, wire_grid, plot_params, 'k')

        tools.draw_gates(ax, self.circuit_diag, labels, gate_grid_index, wire_grid, plot_params, measured,
                         schedule=True)
        plt.show()
        # return ax




"""
-*- coding: utf-8 -*-
@Author : Cui Jinghao
@Time : 2024/6/28 14:43
@Function: quantum simulator.py
@Contact: cuijinghao@tgqs.net
"""

import tgqSim.utils.dev_tools as dev_tools
from tgqSim.circuit.common_gates import (BASE_SINGLE_GATE,
                                               BASE_DOUBLE_GATE,
                                               BASE_TRIPLE_GATE,
                                               MEASURE,
                                               BASE_SINGLE_GATE_MAP,
                                               BASE_DOUBLE_GATE_MAP,
                                               BASE_TRIPLE_GATE_MAP)
from tgqSim.device.noise_models import NOISE_MAPPER, NOISE_TYPE

import tgqSim.device.noise_util as noise_util
from tgqSim.circuit.QuantumCircuit import QuantumCircuit
# import simulator_utils
import os
from typing import Union
import GPUtil
import ctypes
import numpy as np


class Float2(ctypes.Structure):
    _fields_ = [("x", ctypes.c_float),
                ("y", ctypes.c_float)]


class Double2(ctypes.Structure):
    _fields_ = [("x", ctypes.c_double),
                ("y", ctypes.c_double)]


class GateInfo(ctypes.Structure):
    _fields_ = [
        ("gateName", ctypes.c_char_p),
        ("actionPos", ctypes.POINTER(ctypes.c_int)),
        ("theta", ctypes.c_double)
    ]


class SimulationError(Exception):
    def __init__(self, message):
        self.message = message
        super().__init__(self.message)


class QuantumSimulator:

    def __init__(self):
        self.qbit_list = []
        self.state = []
        self.deviceid = []
        self.prob_result = {}

    def set_gpu_device_withlist(self, deviceList: Union[int, list]):
        gpus = GPUtil.getGPUs()
        gpuidList = [gpu.id for gpu in gpus]
        if isinstance(deviceList, int):
            deviceList = [deviceList]
        for deviceid in deviceList:
            if deviceid not in gpuidList:
                raise ValueError("设置设备ID不存在")
        # self.isgpu = True
        # todo: can only run with one kind of device at a time?
        # self.isnpu = False
        self.deviceid = deviceList

    # todo: add a function that does not need devicelist input?
    # todo: expose gpu device list to user?
    def set_gpu_device(self):
        gpus = GPUtil.getGPUs()
        self.deviceid = [gpu.id for gpu in gpus]

    # todo: add set npu device method later
    def set_npu_device(self):
        pass
        # self.isnpu = True
        # self.isgpu = False

    # def validate_obj(self):
    #     if isinstance(self.circuit, QuantumCircuit):
    #         pass

    def _run_with_dcu_device(self, circuit: QuantumCircuit):
        lib = dev_tools.get_dcu_lib()
        lib.execute_circuit.argtypes = [
            ctypes.POINTER(ctypes.POINTER(Double2)),
            ctypes.POINTER(GateInfo),
            ctypes.c_int,
            ctypes.c_int,
            ctypes.POINTER(ctypes.c_int)
        ]
        lib.execute_circuit.restype = None
        gateInfo = []
        for (gate_pos, gate_info) in circuit.gate_list:
            # todo: add an if to check whether gate is measure gate
            if gate_info[0].lower() in MEASURE:
                continue
            elif isinstance(gate_pos, int):
                length = 2
                gate_pos = [gate_pos]
            elif isinstance(gate_pos, list):
                length = len(gate_pos) + 1
            else:
                raise TypeError("Type of gate_pos must be int or list")
            gate_obj = GateInfo()
            actionPos = gate_pos + [-1]
            gate_obj.actionPos = (ctypes.c_int * length)(*actionPos)
            if len(gate_info) > 0:
                gate_obj.gateName = gate_info[0].encode(encoding='utf-8')
            if len(gate_info) > 1:
                gate_obj.theta = gate_info[1]
            gateInfo.append(gate_obj)
        gateInfoCData = (GateInfo * len(gateInfo))(*gateInfo)
        deviceIdCData = (ctypes.c_int * len(self.deviceid))(*self.deviceid)
        # 申请内存首地址，不在Python端申请内存
        # 在C语言中申请统一内存，减少多次拷贝动作
        # todo: separate pointer applying out or not?
        state = ctypes.POINTER(Double2)()
        lib.execute_circuit(ctypes.byref(state), gateInfoCData, len(gateInfo), circuit.width, deviceIdCData)
        # iStart = datetime.datetime.now()
        # print(f"start time is {iStart}")
        state = np.ctypeslib.as_array(state, shape=(2 ** circuit.width,))
        py_state = state.view(np.complex128)
        # clear mem for C++ side
        # lib.freeAllMem(state)
        # print(f"total time of changing type is {(datetime.datetime.now() - iStart).total_seconds()} secs")
        return py_state

    # todo: add run with npu device later
    def _run_with_gpu_device(self, circuit: QuantumCircuit):
        self.set_gpu_device()
        lib = dev_tools.get_cuda_lib()
        lib.execute_circuit.argtypes = [
            ctypes.POINTER(ctypes.POINTER(Double2)),
            ctypes.POINTER(GateInfo),
            ctypes.c_int,
            ctypes.c_int,
            ctypes.POINTER(ctypes.c_int)
        ]
        lib.execute_circuit.restype = None
        gateInfo = []
        for (gate_pos, gate_info) in circuit.gate_list:
            # todo: add an if to check whether gate is measure gate
            if gate_info[0].lower() in MEASURE:
                continue
            elif isinstance(gate_pos, int):
                length = 2
                gate_pos = [gate_pos]
            elif isinstance(gate_pos, list):
                length = len(gate_pos) + 1
            else:
                raise TypeError("Type of gate_pos must be int or list")
            gate_obj = GateInfo()
            actionPos = gate_pos + [-1]
            gate_obj.actionPos = (ctypes.c_int * length)(*actionPos)
            if len(gate_info) > 0:
                gate_obj.gateName = gate_info[0].encode(encoding='utf-8')
            if len(gate_info) > 1:
                gate_obj.theta = gate_info[1]
            gateInfo.append(gate_obj)
        gateInfoCData = (GateInfo * len(gateInfo))(*gateInfo)
        # 以-1作为结尾
        deviceId = self.deviceid + [-1]
        deviceIdCData = (ctypes.c_int * len(deviceId))(*deviceId)
        # 申请内存首地址，不在Python端申请内存
        # 在C语言中申请统一内存，减少多次拷贝动作
        # todo: separate pointer applying out or not?
        state = ctypes.POINTER(Double2)()
        lib.execute_circuit(ctypes.byref(state), gateInfoCData, len(gateInfo), circuit.width, deviceIdCData)
        # iStart = datetime.datetime.now()
        # print(f"start time is {iStart}")
        state = np.ctypeslib.as_array(state, shape=(2 ** circuit.width,))
        py_state = state.view(np.complex128)
        # clear mem for C++ side
        # lib.freeAllMem(state)
        # print(f"total time of changing type is {(datetime.datetime.now() - iStart).total_seconds()} secs")
        return py_state

    # def _run_in_python(self, circuit: QuantumCircuit):

    def _run_with_npu_device(self, circuit: QuantumCircuit):
        import torch
        from tgqSim.NpuGate import SingleGate, DoubleGate, TripleGate
        state = [1 if a == 0 else 0 for a in range(2 ** circuit.width)]
        for (gate_pos, gate_info) in circuit.gate_list:
            # if it is a measure gate, then continue
            if gate_info[0].lower() in MEASURE:
                continue
            gate_type = gate_info[0]
            # last position of gate_info is display_name
            angle = tuple(gate_info[1:])
            if gate_type in BASE_SINGLE_GATE_MAP.keys() or gate_type.upper() in NOISE_MAPPER.values():
                state = SingleGate.ActOn_State(state,
                                                circuit.width,
                                                gate_type,
                                                gate_pos,
                                                *angle)

            elif gate_type in BASE_DOUBLE_GATE_MAP.keys():
                set_gate_pos = set(gate_pos)
                if len(set_gate_pos) != len(gate_pos):
                    raise SimulationError(f"Gate position cannot be the same: {gate_pos[0]}, {gate_pos[1]}")
                state = DoubleGate.ActOn_State(state,
                                                    circuit.width,
                                                    gate_type,
                                                    gate_pos,
                                                    *angle)
            elif gate_type in BASE_TRIPLE_GATE_MAP.keys():
                set_gate_pos = set(gate_pos)
                if len(set_gate_pos) != len(gate_pos):
                    raise SimulationError(f"Gate position cannot be the same: "
                                          f"{gate_pos[0]}, {gate_pos[1]} and {gate_pos[2]}")
                state = TripleGate.ActOn_State(state,
                                                    circuit.width,
                                                    gate_type,
                                                    gate_pos,
                                                    *angle)
            else:
                raise SimulationError(f"Unkown gate type: {gate_type}")
        # tensor to cpu
        state = torch.Tensor.cpu(state)
        return state


    def _run_with_cpu_device(self, circuit: QuantumCircuit):
        from tgqSim.CpuGate import SingleGate, DoubleGate, TripleGate
        # todo: watch out that length of state list grows exponentially with circuit width
        state = [1 if a == 0 else 0 for a in range(2 ** circuit.width)]
        for (gate_pos, gate_info) in circuit.gate_list:
            gate_type = gate_info[0]
            angle = tuple(gate_info[1:])
            # todo: if it is a measure gate, then continue
            if gate_info[0].lower() in MEASURE:
                continue

            # todo: check mapped noise gate
            elif gate_type in BASE_SINGLE_GATE_MAP.keys() or gate_type.upper() in NOISE_MAPPER.values():
                # todo: for noise gate gatetype should coordinate with CpuGate/SingleGate
                state = SingleGate.ActOn_State(state,
                                                circuit.width,
                                                gate_type,
                                                gate_pos,
                                                *angle)
                # if gate_type in NOISE_TYPE:
                #     state = SingleGate.ActOn_State()

            elif gate_type in BASE_DOUBLE_GATE_MAP.keys():
                set_gate_pos = set(gate_pos)
                if len(set_gate_pos) != len(gate_pos):
                    raise SimulationError(f"Gate position cannot be the same: {gate_pos[0]}, {gate_pos[1]}")
                state = DoubleGate.ActOn_State(state,
                                                    circuit.width,
                                                    gate_type,
                                                    gate_pos,
                                                    *angle)
            elif gate_type in BASE_TRIPLE_GATE_MAP.keys():
                set_gate_pos = set(gate_pos)
                if len(set_gate_pos) != len(gate_pos):
                    raise SimulationError(f"Gate position cannot be the same: "
                                          f"{gate_pos[0]}, {gate_pos[1]} and {gate_pos[2]}")
                state = TripleGate.ActOn_State(state,
                                                    circuit.width,
                                                    gate_type,
                                                    gate_pos,
                                                    *angle)
            else:
                raise SimulationError(f"Unkown gate type: {gate_type}")
        return state

    def run_with_noise(self, circuit: QuantumCircuit, shots:int=1000, device: str='cpu'):
        result_dict = {}
        tmp_circuit = circuit.gate_list
        for _ in range(shots):
            new_circuit = []
            for (gate_pos, gate_info) in circuit.noise_circuit:
                # todo: original noise type name
                if gate_info[0] in NOISE_TYPE:
                    # todo: parsed noise type name
                    noise_gate = noise_util.parse_noise(noise_type=gate_info[0], gate_pos=gate_pos, error_rate=gate_info[1])
                    # print(noise_gate)
                    if noise_gate is not None:
                        # new_circuit looks like this [(1, ('h',)), (1, ('amplitude_damp', 0.5))]
                        new_circuit.append(noise_gate)
                else:
                    new_circuit.append((gate_pos, gate_info))
            # print("new_circuit:", new_circuit)
            # todo: check last element of original gate_list, if not measure all bit position
            if tmp_circuit[-1][1][0].lower() == 'measure':
                circuit.gate_list = new_circuit + [(tmp_circuit[-1][0], ('measure', ))]
            else:
                circuit.gate_list = new_circuit
            result = self.execute(circuit, shots=1000, device=device)
            # print(result)
            # print(self.state)
            for key in result.keys():
                if key in result_dict:
                    result_dict[key] += result[key]
                else:
                    result_dict[key] = result[key]
        circuit.gate_list = tmp_circuit

        prob_result = dev_tools.get_normalization(frequency=result_dict)
        # freq_result = self.prob_to_freq(prob_result, shots)
        freq_result = result_dict
        return prob_result, freq_result

    # added npu option
    # todo: add a parameter to identify which device to use
    def run_statevector(self, circuit: QuantumCircuit, device: str = 'cpu'):
        """
        根据线路的门序列计算末态的量子态
        :return:
        """
        if device.upper() == 'GPU':
            return self._run_with_gpu_device(circuit)
        if device.upper() == 'NPU':
            return self._run_with_npu_device(circuit)
        if device.upper() == 'CPU':
            return self._run_with_cpu_device(circuit)
        if device.upper() == 'DCU':
            return self._run_with_dcu_device(circuit)

    @staticmethod
    def freq_to_prob(result: dict):
        shots = sum(result.values())
        prob = {}
        for res, freq in result.items():
            prob[res] = float(freq/shots)
        return prob

    @staticmethod
    def prob_to_freq(result: dict, shots: int):
        freq = {}
        for state, prob in result.items():
            freq[state] = int(prob * shots)
        return freq

    # add device parameter to measure function to identify which device to use
    def execute(self, circuit: QuantumCircuit, shots: int = 1000, device:str='cpu') -> dict:
        """
        execute simulation
        Args:
            circuit:
            # measure_bits_list: 测量比特列表，传入比特位置或列表位置
            shots: 测量次数
            device: cpu, npu or gpu
        Returns:
            result: 返回测量结果
        """
        if shots <= 0:
            raise Exception("Experiment shots cannot be less than or equal to 0!")
        # 首先通过执行操作得到所有的状态
        # print(self.gate_list)
        state = self.run_statevector(circuit, device)
        state = np.array(state)
        # print(state)
        prob = np.real(state.conjugate() * state)
        # e.g. count = {'000': 0, '001': 0, '010': 1, '011': 0, '100': 0, '101': 0, '110': 0, '111': 0}
        count = {format(i, f'0{circuit.width}b'): prob[i] for i in range(len(prob))}
        # print('new count {}'.format(count))
        distribution = {}
        measure_bits_list = None
        # {gatename: position}
        gate_dict = {}
        # (1, ('x',))
        for gate in circuit.gate_list:
            gate_dict.update({gate[1][0]: gate[0]})
        # print(gate_dict)
        if 'measure' in gate_dict.keys():
            # print(gate_dict)
            measure_bits_list = sorted(circuit.width - 1 - i for i in gate_dict['measure'])
        if measure_bits_list is None or measure_bits_list == []:  # 若为空默认测量所有可能
            # bugfix, comment this line
            # distribution = count
            measure_bits_list = sorted([_ for _ in range(circuit.width)])
            # print(measure_bits_list)
        if isinstance(measure_bits_list, int):
            measure_bits_list = [measure_bits_list]

        # print('new measure bit {}'.format(measure_bits_list))
        # 计算测量分布
        for p in count.keys():
            # key = ''.join(p[pos - 1] for pos in measure_bits_list)
            key = ''.join(p[pos] for pos in measure_bits_list)
            if count[p] == 0:
                continue
            if key not in distribution:
                distribution[key] = count[p]
            else:
                distribution[key] += count[p]
        # print('new distribution {}'.format(distribution))
        # 根据分布抽样概率
        result = {}
        cumulate = 0
        sample = np.random.uniform(0, 1, size=shots)
        for key in distribution.keys():
            new_cumulate = cumulate + distribution[key]
            result[key] = sum((cumulate <= sample) & (sample < new_cumulate))
            cumulate = new_cumulate
        return result



"""
-*- coding: utf-8 -*-
@Author : Cui Jinghao
@Time : 2024/8/13 16:39
@Function: setup.py
@Contact: cuijinghao@tgqs.net
"""
from setuptools import setup, find_packages

with open("README.md", "r") as fh:
    long_description = fh.read()

setup(
    name='tgqSim',
    version='1.0.0',
    description='TGQ量子模拟器',  # 包描述
    long_description="Python for quantum simulation http://www.tiangongqs.com",  # 详细描述
    long_description_content_type='text/markdown',
    author='tiangongqs',  # 作者姓名
    license='MIT',  # 许可证
    include_package_data=True,
    packages=find_packages(include=["tgqSim", "tgqSim.*"]),
    install_requires=[
        'numpy>=1.19.0',
        'numba>=0.53.0',
        'matplotlib>=3.3.0',
        'GPUtil>=1.4.0',
        'psutil>=5.9.1',
        'pygments'
    ],
    classifiers=[
    'Operating System :: OS Independent',
    'License :: OSI Approved :: MIT License',
    'Programming Language :: Python :: 3',
    "Programming Language :: Python :: 3.6",
    "Programming Language :: Python :: 3.7",
    "Programming Language :: Python :: 3.8",
    "Programming Language :: Python :: 3.9",
    "Programming Language :: Python :: 3.10",
    "Programming Language :: Python :: 3.11",
    ],
    keywords=["tgq", "quantum", "simulator", "noise"],
)



"""
-*- coding: utf-8 -*-
@Author : XerCis
@Time : 2024/6/28 16:38
@Function:
@Contact: 
"""
import tgqSim.device.noise_models as noise_model
from typing import Union


def parse_noise(noise_type: str, gate_pos: int, error_rate: Union[float, list]):
    """
    对于不同的噪声类型，返回不同的信息

    Args:
        noise_type (str): 噪声类型
        error_rate Union[float, list]: 门错误率信息
    """
    gate_name = "I"
    if "bit_flip" == noise_type:
        gate_name = noise_model.bit_flip(error_rate)
    elif "asymmetric_depolarization" == noise_type:
        gate_name = noise_model.asymmetric_depolarization(error_rate)
    elif "depolarize" == noise_type:
        gate_name = noise_model.depolarize(error_rate)
    elif "phase_flip" == noise_type:
        gate_name = noise_model.phase_flip(error_rate)
    elif "phase_damp" == noise_type:
        gate_name = noise_model.phase_damp(error_rate)
    elif "amplitude_damp" == noise_type:
        gate_name = noise_model.amplitude_damp(error_rate)
    else:
        raise ValueError("Please check noise model name")
    if gate_name == "I":
        return None
    return (gate_pos, (gate_name,error_rate))



"""

@author: Yuchen He
@contact: heyuchen@tgqs.net
@version: 1.0.0
@file: SingleGate.py
@time: 2024/1/16 17:05
"""

import numpy as np
from numba import njit, prange
import math as m
import cmath as cm

def SingleMatirx_ActOn_State(psi, num_qubits, Gate_mat, Gate_pos):
    """
    单比特门矩阵作用于线路的实现
    :param psi:门作用前的状态
    :param num_qubits: 线路总体量子比特数量
    :param Gate_mat: 单比特门矩阵
    :param Gate_pos: 双臂特门矩阵
    :return:
    """
    i = 2 ** Gate_pos
    nstates = 2 ** num_qubits - 2 ** (Gate_pos + 1)
    klist = np.array([k for k in range(0, nstates + 1, i + i)])
    klen = len(klist)
    for k in prange(0, klen):
        for l in prange(0, i):
            i0 = l | klist[k]  # to find a(*...*0_j*...)
            i1 = i0 | i  # to find a(*...*1_j*...) by plus a(00001_j0000)
            temp = np.dot(Gate_mat, np.array([psi[i0], psi[i1]]))
            psi[i0], psi[i1] = temp[0], temp[1]
    return psi

def normalization(psi:list, num_qubits: int):
    norm = np.linalg.norm(psi)
    if 0 == norm:
        return [1 if 0 == i else 0 for i in range(2**num_qubits)]
    else:
        return list(np.divide(psi, norm))


def ActOn_State(psi, num_qubits, gate_type:str, gate_pos, *angles):
    """
    单比特门作用于线路的实现
    :param psi:
    :param num_qubits:
    :param gate_type: 门类型
    :param gate_pos:
    :param angles:
    :return:
    """
    if gate_type.lower() == 'h':
        h_mat = np.array([[1, 1], [1, -1]]) / np.sqrt(2.0)
        return SingleMatirx_ActOn_State(psi, num_qubits, Gate_mat=h_mat, Gate_pos=gate_pos)
    elif gate_type.lower() == 'x':
        x_mat = np.array([[0, 1], [1, 0]])
        return SingleMatirx_ActOn_State(psi, num_qubits, Gate_mat=x_mat, Gate_pos=gate_pos)
    elif gate_type.lower() == 'y':
        y_mat = np.array([[0, -1j], [1j, 0]])
        return SingleMatirx_ActOn_State(psi, num_qubits, Gate_mat=y_mat, Gate_pos=gate_pos)
    elif gate_type.lower() == 'z':
        z_mat = np.array([[1,0],[0,-1]])
        return SingleMatirx_ActOn_State(psi, num_qubits, Gate_mat=z_mat, Gate_pos=gate_pos)
    elif gate_type.lower() == 'u3':
        u3_mat = np.array([
            [m.cos(angles[0] / 2.0), -cm.exp(1j * angles[2]) * m.sin(angles[0] / 2.0)],
            [cm.exp(1j * angles[1]) * m.sin(angles[0] / 2.0),
             cm.exp(1j * (angles[1] + angles[2])) * m.cos(angles[0] / 2.0)]
        ])
        return SingleMatirx_ActOn_State(psi, num_qubits, Gate_mat=u3_mat, Gate_pos=gate_pos)
    elif gate_type.lower() == 'rx':
        rx_mat = np.array([[m.cos(angles[0] / 2.0), -1j * m.sin(angles[0] / 2.0)],
                           [-1j * m.sin(angles[0] / 2.0), m.cos(angles[0] / 2.0)]]
                          )
        return SingleMatirx_ActOn_State(psi, num_qubits, Gate_mat=rx_mat, Gate_pos=gate_pos)
    elif gate_type.lower() == 'ry':
        ry_mat = np.array([[m.cos(angles[0] / 2.0), -m.sin(angles[0] / 2.0)],
                           [m.sin(angles[0] / 2.0), m.cos(angles[0] / 2.0)]]
                          )
        return SingleMatirx_ActOn_State(psi, num_qubits, Gate_mat=ry_mat, Gate_pos=gate_pos)
    elif gate_type.lower() == 'rz':
        rz_mat = np.array([[cm.exp(-1j * angles[0] / 2.0), 0],
                           [0, cm.exp(1j * angles[0] / 2.0)]])
        return SingleMatirx_ActOn_State(psi, num_qubits, Gate_mat=rz_mat, Gate_pos=gate_pos)
    elif gate_type.lower() == 's':
        rz_mat = np.array(
            [
                [1, 0],
                [0, 1j]
            ]
        )
        return SingleMatirx_ActOn_State(psi, num_qubits, Gate_mat=rz_mat, Gate_pos=gate_pos)
    elif gate_type.lower() == 'sdg':
        rz_mat = np.array(
            [
                [1, 0],
                [0, -1j]
            ]
        )
        return SingleMatirx_ActOn_State(psi, num_qubits, Gate_mat=rz_mat, Gate_pos=gate_pos)
    elif gate_type.lower() == 't':
        rz_mat = np.array(
            [
                [1, 0], 
                [0, cm.exp(np.pi / 4 * 1j)]
            ]
        )
        return SingleMatirx_ActOn_State(psi, num_qubits, Gate_mat=rz_mat, Gate_pos=gate_pos)
    elif gate_type.lower() == 'tdg':
        rz_mat = np.array(
            [
                [1, 0], 
                [0, cm.exp(-np.pi / 4 * 1j)]
            ]
        )
        return SingleMatirx_ActOn_State(psi, num_qubits, Gate_mat=rz_mat, Gate_pos=gate_pos)
    elif gate_type.lower() == "damp_i":
        error_rate = angles[0]
        rz_mat = np.array(
            [
                [1, 0],
                [0, np.sqrt(1 - error_rate)]
            ]
        )
        return normalization(SingleMatirx_ActOn_State(psi, num_qubits, Gate_mat=rz_mat, Gate_pos=gate_pos), num_qubits)
    elif gate_type.lower() == "pd":
        error_rate = angles[0]
        rz_mat = np.array(
            [
                [0, 0],
                [0, np.sqrt(error_rate)]
            ]
        )
        return normalization(SingleMatirx_ActOn_State(psi, num_qubits, Gate_mat=rz_mat, Gate_pos=gate_pos), num_qubits)
    elif gate_type.lower() == "ad":
        error_rate = angles[0]
        rz_mat = np.array(
            [
                [0, np.sqrt(error_rate)],
                [0, 0]
            ]
        )
        return normalization(SingleMatirx_ActOn_State(psi, num_qubits, Gate_mat=rz_mat, Gate_pos=gate_pos), num_qubits)


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
@author: Yuchen He
@contact: heyuchen@tgqs.net
@version: 1.0.0
@file: TripleGate.py
@time: 2024/1/16 17:05
"""
from numba import njit, prange
def ActOn_State(psi, num_qubits, Gate_type, Gate_pos, *Angles):
    """

    :param psi:
    :param num_qubits:
    :param Gate_type:
    :param Gate_pos:
    :param Angles:
    :return:
    """

    # todo 目前只有Toffoli门
    j2 = sorted(Gate_pos)[2]  # 高位比特，不区分控制位与目标位
    j1 = sorted(Gate_pos)[1]  # 中间位比特，不区分控制位与目标位
    j0 = sorted(Gate_pos)[0]  # 低位比特，不区分控制位与目标位
    # j0<j1<j2
    i1_plus = 2 ** j0
    i2_plus = 2 ** j1
    i3_plus = 2 ** j0 + 2 ** j1
    i4_plus = 2 ** j2
    i5_plus = 2 ** j2 + 2 ** j0
    i6_plus = 2 ** j2 + 2 ** j1
    i7_plus = 2 ** j2 + 2 ** j1 + 2 ** j0
    delta3 = 2 ** (j2 + 1)
    delta2 = 2 ** (j1 + 1)
    delta1 = 2 ** (j0 + 1)
    max3 = 2 ** num_qubits - 2 ** (j2 + 1)
    max2 = 2 ** j2 - 2 ** (j1 + 1)
    max1 = 2 ** j1 - 2 ** (j0 + 1)
    max0 = 2 ** j0 - 1

    for k in prange(0, max3 + 1, delta3):
        for l in prange(0, max2 + 1, delta2):
            for m in prange(0, max1 + 1, delta1):
                for n in prange(0, max0 + 1):
                    i0 = m | l | k | n  # to get index of a(*...*0_j2*...*0_j1*...*0_j0*...*)
                    i1 = i0 | i1_plus  # to get index of a(*...*0_j2*...*0_j1*...*1_j0*...*)
                    i2 = i0 | i2_plus  # to get index of a(*...*0_j2*...*1_j1*...*0_j0*...*)
                    i3 = i0 | i3_plus  # to get index of a(*...*0_j2*...*1_j1*...*1_j0*...*)
                    i4 = i0 | i4_plus  # to get index of a(*...*1_j2*...*0_j1*...*0_j0*...*)
                    i5 = i0 | i5_plus  # to get index of a(*...*1_j2*...*0_j1*...*1_j0*...*)
                    i6 = i0 | i6_plus  # to get index of a(*...*1_j2*...*1_j1*...*0_j0*...*)
                    i7 = i0 | i7_plus  # to get index of a(*...*1_j2*...*1_j1*...*1_j0*...*)
                    if Gate_type == 'ccx':
                        # use rule to update matrix element of state
                        if Gate_pos[0] < Gate_pos[2] and Gate_pos[1] < Gate_pos[2]:
                            # case: control_q: q1,q0(q0,q1), target_q: q2 
                            # 011 -> 111 , 111 -> 011
                            psi[i3], psi[i7] = psi[i7], psi[i3]

                        elif Gate_pos[2] < Gate_pos[0] and Gate_pos[2] < Gate_pos[1]:
                            # case: control_q: q2q1(q1q2), target_q: q0
                            # 110 -> 111, 111-> 110
                            psi[i6], psi[i7] = psi[i7], psi[i6]

                        elif (Gate_pos[2] > Gate_pos[0] and Gate_pos[2] < Gate_pos[1]) or (
                                Gate_pos[2] > Gate_pos[1] and Gate_pos[2] < Gate_pos[0]):
                            # case: control_q: q0q2(q2q0), target_q: q1
                            # 101 -> 111 , 111 -> 101
                            psi[i5], psi[i7] = psi[i7], psi[i5]
                    elif Gate_type == "cswap":
                        if Gate_pos[0] < min(Gate_pos[1], Gate_pos[2]):
                            psi[i3], psi[i5] = psi[i5], psi[i3]
                        elif Gate_pos[1] < min(Gate_pos[0], Gate_pos[2]):
                            psi[i3], psi[i6] = psi[i6], psi[i3]
                        elif Gate_pos[2] < min(Gate_pos[0], Gate_pos[1]):
                            psi[i6], psi[i5] = psi[i5], psi[i6]
    return psi



"""
-*- coding: utf-8 -*-
@Author : XerCis
@Time : 2024/7/2 15:00
@Function:
@Contact: 
"""


def setdevice(self, deviceList: Union[int, list]):
    gpus = GPUtil.getGPUs()
    gpuidList = [gpu.id for gpu in gpus]
    if isinstance(deviceList, int):
        deviceList = [deviceList]
    for deviceid in deviceList:
        if deviceid not in gpuidList:
            raise ValueError("设置设备ID不存在")
    self.isgpu = True
    self.deviceid = deviceList



"""
-*- coding: utf-8 -*-
@Author : Cui Jinghao
@Time : 2024/9/11 15:40
@Function: common_decompositions.py
@Contact: cuijinghao@tgqs.net
"""


import tgqSim as tgqs
import numpy as np

# sqrt_x = tgqs.X**0.5
'''
对不同门拆解进行初始化，拆解线路的基础门集采用{Rz, cz, X, SX},以下只选取了一种拆解方式，对不同拆解方式可自行定义。
'''



def H(qubit, width):
    circuit = tgqs.QuantumCircuit(width)
    circuit.append(tgqs.rz(qubit, np.pi/2))
    circuit.append(tgqs.sqrt_x(qubit))
    circuit.append(tgqs.rz(qubit, np.pi/2))
    return circuit


def Y(qubit, width):
    circuit = tgqs.QuantumCircuit(width)
    circuit.append(tgqs.rz(qubit, -np.pi))
    circuit.append(tgqs.x(qubit))
    return circuit


def Z(qubit, width):
    circuit = tgqs.QuantumCircuit(width)
    circuit.append(tgqs.rz(qubit, np.pi))
    return circuit

def Cnot(control_qubit, target_qubit, width):
    circuit = tgqs.QuantumCircuit(width)
    circuit.append(tgqs.rz(target_qubit, np.pi/2))
    circuit.append(tgqs.sqrt_x(target_qubit))
    circuit.append(tgqs.rz(target_qubit, np.pi/2))
    circuit.append(tgqs.cz(control_qubit, target_qubit))
    circuit.append(tgqs.rz(target_qubit, np.pi/2))
    circuit.append(tgqs.sqrt_x(target_qubit))
    circuit.append(tgqs.rz(target_qubit, np.pi/2))
    return circuit

def S(qubit, width):
    circuit = tgqs.QuantumCircuit(width)
    circuit.append(tgqs.rz(qubit, np.pi/2))
    return circuit

def Sdg(qubit, width):
    circuit = tgqs.QuantumCircuit(width)
    circuit.append(tgqs.rz(qubit, -np.pi/2))
    return circuit

def T(qubit, width):
    circuit = tgqs.QuantumCircuit(width)
    circuit.append(tgqs.rz(qubit, np.pi/4))
    return circuit

def Tdg(qubit, width):
    circuit = tgqs.QuantumCircuit(width)
    circuit.append(tgqs.rz(qubit, -np.pi/4))
    return circuit

def Ch(control_qubit, target_qubit, width):
    circuit = tgqs.QuantumCircuit(width)
    circuit.append(tgqs.rz(control_qubit, np.pi))
    circuit.append(tgqs.sqrt_x(control_qubit))
    circuit.append(tgqs.rz(control_qubit, np.pi*3/4))
    # circuit.append(tgqs.CNOT(target_qubit, control_qubit))
    circuit.append(tgqs.rz(target_qubit, np.pi/2))
    circuit.append(tgqs.sqrt_x(target_qubit))
    circuit.append(tgqs.rz(target_qubit, np.pi/2))
    circuit.append(tgqs.cz(control_qubit, target_qubit))
    circuit.append(tgqs.rz(target_qubit, np.pi/2))
    circuit.append(tgqs.sqrt_x(target_qubit))
    circuit.append(tgqs.rz(target_qubit, np.pi/2))
    circuit.append(tgqs.rz(control_qubit, np.pi/4))
    circuit.append(tgqs.sqrt_x(control_qubit))
    return circuit

def Rx(theta, qubit, width):
    circuit = tgqs.QuantumCircuit(width)
    circuit.append(tgqs.rz(qubit, np.pi/2))
    circuit.append(tgqs.sqrt_x(qubit))
    circuit.append(tgqs.rz(qubit, np.pi+theta))
    circuit.append(tgqs.sqrt_x(qubit))
    circuit.append(tgqs.rz(qubit, np.pi*5/2))
    return circuit

def Ry(theta, qubit, width):
    circuit = tgqs.QuantumCircuit(width)
    circuit.append(tgqs.rz(qubit, 0))
    circuit.append(tgqs.sqrt_x(qubit))
    circuit.append(tgqs.rz(qubit, np.pi+theta))
    circuit.append(tgqs.sqrt_x(qubit))
    circuit.append(tgqs.rz(qubit, np.pi*3))
    return circuit

def Swap(qubit1, qubit2, width):
    circuit = tgqs.QuantumCircuit(width)
    # circuit.append(tgqs.CNOT(qubit1, qubit2))
    # circuit.append(tgqs.CNOT(qubit2, qubit1))
    # circuit.append(tgqs.CNOT(qubit1, qubit2))
    circuit.append(tgqs.rz(qubit2, np.pi / 2))
    circuit.append(tgqs.sqrt_x(qubit2))
    circuit.append(tgqs.rz(qubit2, np.pi / 2))
    circuit.append(tgqs.cz(qubit1, qubit2))
    circuit.append(tgqs.rz(qubit2, np.pi / 2))
    circuit.append(tgqs.sqrt_x(qubit2))
    circuit.append(tgqs.rz(qubit2, np.pi / 2))

    circuit.append(tgqs.rz(qubit1, np.pi / 2))
    circuit.append(tgqs.sqrt_x(qubit1))
    circuit.append(tgqs.rz(qubit1, np.pi / 2))
    circuit.append(tgqs.cz(qubit2, qubit1))
    circuit.append(tgqs.rz(qubit1, np.pi / 2))
    circuit.append(tgqs.sqrt_x(qubit1))
    circuit.append(tgqs.rz(qubit1, np.pi / 2))

    circuit.append(tgqs.rz(qubit2, np.pi / 2))
    circuit.append(tgqs.sqrt_x(qubit2))
    circuit.append(tgqs.rz(qubit2, np.pi / 2))
    circuit.append(tgqs.cz(qubit1, qubit2))
    circuit.append(tgqs.rz(qubit2, np.pi / 2))
    circuit.append(tgqs.sqrt_x(qubit2))
    circuit.append(tgqs.rz(qubit2, np.pi / 2))
    return circuit

def ISwap(qubit1, qubit2, width):
    circuit = tgqs.QuantumCircuit(width)
    circuit.append(tgqs.rz(qubit1, np.pi))
    circuit.append(tgqs.rz(qubit2, np.pi/2))
    circuit.append(tgqs.sqrt_x(qubit1))
    circuit.append(tgqs.rz(qubit1, np.pi/2))
    # circuit.append(tgqs.CNOT(qubit1, qubit2))
    # circuit.append(tgqs.CNOT(qubit2, qubit1))
    circuit.append(tgqs.rz(qubit2, np.pi / 2))
    circuit.append(tgqs.sqrt_x(qubit2))
    circuit.append(tgqs.rz(qubit2, np.pi / 2))
    circuit.append(tgqs.cz(qubit1, qubit2))
    circuit.append(tgqs.rz(qubit2, np.pi / 2))
    circuit.append(tgqs.sqrt_x(qubit2))
    circuit.append(tgqs.rz(qubit2, np.pi / 2))

    circuit.append(tgqs.rz(qubit1, np.pi / 2))
    circuit.append(tgqs.sqrt_x(qubit1))
    circuit.append(tgqs.rz(qubit1, np.pi / 2))
    circuit.append(tgqs.cz(qubit2, qubit1))
    circuit.append(tgqs.rz(qubit1, np.pi / 2))
    circuit.append(tgqs.sqrt_x(qubit1))
    circuit.append(tgqs.rz(qubit1, np.pi / 2))
    circuit.append(tgqs.rz(qubit2, np.pi/2))
    circuit.append(tgqs.sqrt_x(qubit2))
    circuit.append(tgqs.rz(qubit2, np.pi/2))
    return circuit

def Cphase(theta, control_qubit, target_qubit, width):
    circuit = tgqs.QuantumCircuit(width)
    circuit.append(tgqs.rz(control_qubit, theta/2))
    # circuit.append(tgqs.CNOT(control_qubit, target_qubit))
    circuit.append(tgqs.rz(target_qubit, np.pi/2))
    circuit.append(tgqs.sqrt_x(target_qubit))
    circuit.append(tgqs.rz(target_qubit, np.pi/2))
    circuit.append(tgqs.cz(control_qubit, target_qubit))
    circuit.append(tgqs.rz(target_qubit, np.pi/2))
    circuit.append(tgqs.sqrt_x(target_qubit))
    circuit.append(tgqs.rz(target_qubit, np.pi/2))
    circuit.append(tgqs.rz(target_qubit, -(theta/2)))
    # circuit.append(tgqs.CNOT(control_qubit, target_qubit))
    circuit.append(tgqs.rz(target_qubit, np.pi/2))
    circuit.append(tgqs.sqrt_x(target_qubit))
    circuit.append(tgqs.rz(target_qubit, np.pi/2))
    circuit.append(tgqs.cz(control_qubit, target_qubit))
    circuit.append(tgqs.rz(target_qubit, np.pi/2))
    circuit.append(tgqs.sqrt_x(target_qubit))
    circuit.append(tgqs.rz(target_qubit, np.pi/2))
    circuit.append(tgqs.rz(target_qubit, theta/2))
    return circuit

def Fred(control_qubit, qubit1, qubit2, width):
    circuit = tgqs.QuantumCircuit(width)
    # circuit.append(tgqs.CNOT(qubit1, control_qubit))
    circuit.append(tgqs.rz(control_qubit, np.pi / 2))
    circuit.append(tgqs.sqrt_x(control_qubit))
    circuit.append(tgqs.rz(control_qubit, np.pi / 2))
    circuit.append(tgqs.cz(qubit1, control_qubit))
    circuit.append(tgqs.rz(control_qubit, np.pi / 2))
    circuit.append(tgqs.sqrt_x(control_qubit))
    circuit.append(tgqs.rz(control_qubit, np.pi / 2))

    circuit.append(tgqs.rz(qubit1, np.pi/2))
    circuit.append(tgqs.sqrt_x(qubit1))
    circuit.append(tgqs.rz(qubit1, np.pi/2))
    # circuit.append(tgqs.CNOT(control_qubit, qubit1))
    circuit.append(tgqs.rz(qubit1, np.pi / 2))
    circuit.append(tgqs.sqrt_x(qubit1))
    circuit.append(tgqs.rz(qubit1, np.pi / 2))
    circuit.append(tgqs.cz(control_qubit, qubit1))
    circuit.append(tgqs.rz(qubit1, np.pi / 2))
    circuit.append(tgqs.sqrt_x(qubit1))
    circuit.append(tgqs.rz(qubit1, np.pi / 2))

    circuit.append(tgqs.rz(qubit1, -(np.pi/4)))
    # circuit.append(tgqs.CNOT(qubit2, qubit1))
    circuit.append(tgqs.rz(qubit1, np.pi / 2))
    circuit.append(tgqs.sqrt_x(qubit1))
    circuit.append(tgqs.rz(qubit1, np.pi / 2))
    circuit.append(tgqs.cz(qubit2, qubit1))
    circuit.append(tgqs.rz(qubit1, np.pi / 2))
    circuit.append(tgqs.sqrt_x(qubit1))
    circuit.append(tgqs.rz(qubit1, np.pi / 2))

    circuit.append(tgqs.rz(qubit1, np.pi/4))
    # circuit.append(tgqs.CNOT(control_qubit, qubit1))
    circuit.append(tgqs.rz(qubit1, np.pi / 2))
    circuit.append(tgqs.sqrt_x(qubit1))
    circuit.append(tgqs.rz(qubit1, np.pi / 2))
    circuit.append(tgqs.cz(control_qubit, qubit1))
    circuit.append(tgqs.rz(qubit1, np.pi / 2))
    circuit.append(tgqs.sqrt_x(qubit1))
    circuit.append(tgqs.rz(qubit1, np.pi / 2))

    circuit.append(tgqs.rz(control_qubit, np.pi/4))
    circuit.append(tgqs.rz(qubit1, -(np.pi/4)))
    # circuit.append(tgqs.CNOT(qubit2, qubit1))
    circuit.append(tgqs.rz(qubit1, np.pi / 2))
    circuit.append(tgqs.sqrt_x(qubit1))
    circuit.append(tgqs.rz(qubit1, np.pi / 2))
    circuit.append(tgqs.cz(qubit2, qubit1))
    circuit.append(tgqs.rz(qubit1, np.pi / 2))
    circuit.append(tgqs.sqrt_x(qubit1))
    circuit.append(tgqs.rz(qubit1, np.pi / 2))

    circuit.append(tgqs.rz(qubit1, np.pi*3/4))
    circuit.append(tgqs.sqrt_x(qubit1))
    circuit.append(tgqs.rz(qubit1, np.pi/2))
    # circuit.append(tgqs.CNOT(control_qubit, qubit1))
    # circuit.append(tgqs.CNOT(qubit1, control_qubit))
    # circuit.append(tgqs.CNOT(control_qubit, qubit1))
    circuit.append(tgqs.rz(qubit1, np.pi / 2))
    circuit.append(tgqs.sqrt_x(qubit1))
    circuit.append(tgqs.rz(qubit1, np.pi / 2))
    circuit.append(tgqs.cz(control_qubit, qubit1))
    circuit.append(tgqs.rz(qubit1, np.pi / 2))
    circuit.append(tgqs.sqrt_x(qubit1))
    circuit.append(tgqs.rz(qubit1, np.pi / 2))

    circuit.append(tgqs.rz(control_qubit, np.pi / 2))
    circuit.append(tgqs.sqrt_x(control_qubit))
    circuit.append(tgqs.rz(control_qubit, np.pi / 2))
    circuit.append(tgqs.cz(qubit1, control_qubit))
    circuit.append(tgqs.rz(control_qubit, np.pi / 2))
    circuit.append(tgqs.sqrt_x(control_qubit))
    circuit.append(tgqs.rz(control_qubit, np.pi / 2))

    circuit.append(tgqs.rz(qubit1, np.pi / 2))
    circuit.append(tgqs.sqrt_x(qubit1))
    circuit.append(tgqs.rz(qubit1, np.pi / 2))
    circuit.append(tgqs.cz(control_qubit, qubit1))
    circuit.append(tgqs.rz(qubit1, np.pi / 2))
    circuit.append(tgqs.sqrt_x(qubit1))
    circuit.append(tgqs.rz(qubit1, np.pi / 2))

    # circuit.append(tgqs.CNOT(qubit2, qubit1))
    circuit.append(tgqs.rz(qubit1, np.pi / 2))
    circuit.append(tgqs.sqrt_x(qubit1))
    circuit.append(tgqs.rz(qubit1, np.pi / 2))
    circuit.append(tgqs.cz(qubit2, qubit1))
    circuit.append(tgqs.rz(qubit1, np.pi / 2))
    circuit.append(tgqs.sqrt_x(qubit1))
    circuit.append(tgqs.rz(qubit1, np.pi / 2))
    circuit.append(tgqs.rz(qubit1, -(np.pi/4)))
    circuit.append(tgqs.rz(qubit2, np.pi/4))
    # circuit.append(tgqs.CNOT(qubit2, qubit1))
    # circuit.append(tgqs.CNOT(control_qubit, qubit1))
    circuit.append(tgqs.rz(qubit1, np.pi / 2))
    circuit.append(tgqs.sqrt_x(qubit1))
    circuit.append(tgqs.rz(qubit1, np.pi / 2))
    circuit.append(tgqs.cz(qubit2, qubit1))
    circuit.append(tgqs.rz(qubit1, np.pi / 2))
    circuit.append(tgqs.sqrt_x(qubit1))
    circuit.append(tgqs.rz(qubit1, np.pi / 2))

    circuit.append(tgqs.rz(qubit1, np.pi / 2))
    circuit.append(tgqs.sqrt_x(qubit1))
    circuit.append(tgqs.rz(qubit1, np.pi / 2))
    circuit.append(tgqs.cz(control_qubit, qubit1))
    circuit.append(tgqs.rz(qubit1, np.pi / 2))
    circuit.append(tgqs.sqrt_x(qubit1))
    circuit.append(tgqs.rz(qubit1, np.pi / 2))
    return circuit

def Syc(qubit1, qubit2, width):
    circuit = tgqs.QuantumCircuit(width)
    circuit.append(tgqs.rz(qubit1, -np.pi))
    circuit.append(tgqs.rz(qubit2, -np.pi/2))
    circuit.append(tgqs.sqrt_x(qubit1))
    circuit.append(tgqs.sqrt_x(qubit2))
    circuit.append(tgqs.rz(qubit1, np.pi/2))
    circuit.append(tgqs.rz(qubit2, -np.pi))
    # circuit.append(tgqs.CNOT(qubit1, qubit2))
    circuit.append(tgqs.rz(qubit2, np.pi / 2))
    circuit.append(tgqs.sqrt_x(qubit2))
    circuit.append(tgqs.rz(qubit2, np.pi / 2))
    circuit.append(tgqs.cz(qubit1, qubit2))
    circuit.append(tgqs.rz(qubit2, np.pi / 2))
    circuit.append(tgqs.sqrt_x(qubit2))
    circuit.append(tgqs.rz(qubit2, np.pi / 2))

    circuit.append(tgqs.sqrt_x(qubit1))
    circuit.append(tgqs.rz(qubit2, np.pi/2))
    circuit.append(tgqs.rz(qubit1, np.pi))
    # circuit.append(tgqs.CNOT(qubit1, qubit2))
    circuit.append(tgqs.rz(qubit2, np.pi / 2))
    circuit.append(tgqs.sqrt_x(qubit2))
    circuit.append(tgqs.rz(qubit2, np.pi / 2))
    circuit.append(tgqs.cz(qubit1, qubit2))
    circuit.append(tgqs.rz(qubit2, np.pi / 2))
    circuit.append(tgqs.sqrt_x(qubit2))
    circuit.append(tgqs.rz(qubit2, np.pi / 2))

    circuit.append(tgqs.sqrt_x(qubit1))
    circuit.append(tgqs.rz(qubit2, np.pi/2))
    # circuit.append(tgqs.CNOT(qubit1, qubit2))
    circuit.append(tgqs.rz(qubit2, np.pi / 2))
    circuit.append(tgqs.sqrt_x(qubit2))
    circuit.append(tgqs.rz(qubit2, np.pi / 2))
    circuit.append(tgqs.cz(qubit1, qubit2))
    circuit.append(tgqs.rz(qubit2, np.pi / 2))
    circuit.append(tgqs.sqrt_x(qubit2))
    circuit.append(tgqs.rz(qubit2, np.pi / 2))
    circuit.append(tgqs.x(qubit1))
    circuit.append(tgqs.x(qubit2))
    circuit.append(tgqs.rz(qubit1, np.pi*11/12))
    circuit.append(tgqs.rz(qubit2, np.pi*11/12))
    return circuit

def Rxx(theta, qubit1, qubit2, width):
    circuit = tgqs.QuantumCircuit(width)
    circuit.append(tgqs.rz(qubit1, np.pi/2))
    circuit.append(tgqs.rz(qubit2, np.pi/2))
    circuit.append(tgqs.sqrt_x(qubit1))
    circuit.append(tgqs.sqrt_x(qubit2))
    circuit.append(tgqs.rz(qubit1, np.pi/2))
    circuit.append(tgqs.rz(qubit2, np.pi))
    # circuit.append(tgqs.CNOT(qubit2, qubit1))
    circuit.append(tgqs.rz(qubit1, np.pi / 2))
    circuit.append(tgqs.sqrt_x(qubit1))
    circuit.append(tgqs.rz(qubit1, np.pi / 2))
    circuit.append(tgqs.cz(qubit2, qubit1))
    circuit.append(tgqs.rz(qubit1, np.pi / 2))
    circuit.append(tgqs.sqrt_x(qubit1))
    circuit.append(tgqs.rz(qubit1, np.pi / 2))

    circuit.append(tgqs.rz(qubit1, theta))
    # circuit.append(tgqs.CNOT(qubit2, qubit1))
    circuit.append(tgqs.rz(qubit1, np.pi / 2))
    circuit.append(tgqs.sqrt_x(qubit1))
    circuit.append(tgqs.rz(qubit1, np.pi / 2))
    circuit.append(tgqs.cz(qubit2, qubit1))
    circuit.append(tgqs.rz(qubit1, np.pi / 2))
    circuit.append(tgqs.sqrt_x(qubit1))
    circuit.append(tgqs.rz(qubit1, np.pi / 2))

    circuit.append(tgqs.rz(qubit1, np.pi/2))
    circuit.append(tgqs.sqrt_x(qubit2))
    circuit.append(tgqs.sqrt_x(qubit1))
    circuit.append(tgqs.rz(qubit2, np.pi/2))
    circuit.append(tgqs.rz(qubit1, np.pi/2))
    return circuit

def Ryy(theta, qubit1, qubit2, width):
    circuit = tgqs.QuantumCircuit(width)
    circuit.append(tgqs.sqrt_x(qubit1))
    circuit.append(tgqs.sqrt_x(qubit2))
    # circuit.append(tgqs.CNOT(qubit1, qubit2))
    circuit.append(tgqs.rz(qubit2, np.pi / 2))
    circuit.append(tgqs.sqrt_x(qubit2))
    circuit.append(tgqs.rz(qubit2, np.pi / 2))
    circuit.append(tgqs.cz(qubit1, qubit2))
    circuit.append(tgqs.rz(qubit2, np.pi / 2))
    circuit.append(tgqs.sqrt_x(qubit2))
    circuit.append(tgqs.rz(qubit2, np.pi / 2))

    circuit.append(tgqs.rz(qubit2, theta))
    # circuit.append(tgqs.CNOT(qubit1, qubit2))
    circuit.append(tgqs.rz(qubit2, np.pi / 2))
    circuit.append(tgqs.sqrt_x(qubit2))
    circuit.append(tgqs.rz(qubit2, np.pi / 2))
    circuit.append(tgqs.cz(qubit1, qubit2))
    circuit.append(tgqs.rz(qubit2, np.pi / 2))
    circuit.append(tgqs.sqrt_x(qubit2))
    circuit.append(tgqs.rz(qubit2, np.pi / 2))

    circuit.append(tgqs.rz(qubit1, -np.pi))
    circuit.append(tgqs.rz(qubit2, -np.pi))
    circuit.append(tgqs.sqrt_x(qubit1))
    circuit.append(tgqs.sqrt_x(qubit2))
    circuit.append(tgqs.rz(qubit1, -np.pi))
    circuit.append(tgqs.rz(qubit2, -np.pi))
    return circuit

def Rzz(theta, qubit1, qubit2, width):
    circuit = tgqs.QuantumCircuit(width)
    # circuit.append(tgqs.CNOT(qubit1, qubit2))
    circuit.append(tgqs.rz(qubit2, np.pi / 2))
    circuit.append(tgqs.sqrt_x(qubit2))
    circuit.append(tgqs.rz(qubit2, np.pi / 2))
    circuit.append(tgqs.cz(qubit1, qubit2))
    circuit.append(tgqs.rz(qubit2, np.pi / 2))
    circuit.append(tgqs.sqrt_x(qubit2))
    circuit.append(tgqs.rz(qubit2, np.pi / 2))
    circuit.append(tgqs.rz(qubit2, theta))
    # circuit.append(tgqs.CNOT(qubit1, qubit2))
    circuit.append(tgqs.rz(qubit2, np.pi / 2))
    circuit.append(tgqs.sqrt_x(qubit2))
    circuit.append(tgqs.rz(qubit2, np.pi / 2))
    circuit.append(tgqs.cz(qubit1, qubit2))
    circuit.append(tgqs.rz(qubit2, np.pi / 2))
    circuit.append(tgqs.sqrt_x(qubit2))
    circuit.append(tgqs.rz(qubit2, np.pi / 2))
    return circuit

def Toff(qubit1, qubit2, qubit3, width):
    circuit = tgqs.QuantumCircuit(width)
    circuit.append(tgqs.rz(qubit2, np.pi/2))
    circuit.append(tgqs.sqrt_x(qubit2))
    circuit.append(tgqs.rz(qubit2, np.pi/2))
    # circuit.append(tgqs.CNOT(qubit1, qubit2))
    circuit.append(tgqs.rz(qubit2, np.pi / 2))
    circuit.append(tgqs.sqrt_x(qubit2))
    circuit.append(tgqs.rz(qubit2, np.pi / 2))
    circuit.append(tgqs.cz(qubit1, qubit2))
    circuit.append(tgqs.rz(qubit2, np.pi / 2))
    circuit.append(tgqs.sqrt_x(qubit2))
    circuit.append(tgqs.rz(qubit2, np.pi / 2))

    circuit.append(tgqs.rz(qubit2, -np.pi/4))
    # circuit.append(tgqs.CNOT(qubit3, qubit2))
    circuit.append(tgqs.rz(qubit2, np.pi / 2))
    circuit.append(tgqs.sqrt_x(qubit2))
    circuit.append(tgqs.rz(qubit2, np.pi / 2))
    circuit.append(tgqs.cz(qubit3, qubit2))
    circuit.append(tgqs.rz(qubit2, np.pi / 2))
    circuit.append(tgqs.sqrt_x(qubit2))
    circuit.append(tgqs.rz(qubit2, np.pi / 2))

    circuit.append(tgqs.rz(qubit2, np.pi/4))
    # circuit.append(tgqs.CNOT(qubit1, qubit2))
    circuit.append(tgqs.rz(qubit2, np.pi / 2))
    circuit.append(tgqs.sqrt_x(qubit2))
    circuit.append(tgqs.rz(qubit2, np.pi / 2))
    circuit.append(tgqs.cz(qubit1, qubit2))
    circuit.append(tgqs.rz(qubit2, np.pi / 2))
    circuit.append(tgqs.sqrt_x(qubit2))
    circuit.append(tgqs.rz(qubit2, np.pi / 2))

    circuit.append(tgqs.rz(qubit1, np.pi/4))
    circuit.append(tgqs.rz(qubit2, -np.pi/4))
    # circuit.append(tgqs.CNOT(qubit3, qubit2))
    circuit.append(tgqs.rz(qubit2, np.pi / 2))
    circuit.append(tgqs.sqrt_x(qubit2))
    circuit.append(tgqs.rz(qubit2, np.pi / 2))
    circuit.append(tgqs.cz(qubit3, qubit2))
    circuit.append(tgqs.rz(qubit2, np.pi / 2))
    circuit.append(tgqs.sqrt_x(qubit2))
    circuit.append(tgqs.rz(qubit2, np.pi / 2))

    circuit.append(tgqs.rz(qubit2, np.pi*3/4))
    circuit.append(tgqs.sqrt_x(qubit2))
    circuit.append(tgqs.rz(qubit2, np.pi/2))
    # circuit.append(tgqs.CNOT(qubit3, qubit2))
    # circuit.append(tgqs.CNOT(qubit2, qubit3))
    # circuit.append(tgqs.CNOT(qubit3, qubit2))
    # circuit.append(tgqs.CNOT(qubit2, qubit1))
    circuit.append(tgqs.rz(qubit2, np.pi / 2))
    circuit.append(tgqs.sqrt_x(qubit2))
    circuit.append(tgqs.rz(qubit2, np.pi / 2))
    circuit.append(tgqs.cz(qubit3, qubit2))
    circuit.append(tgqs.rz(qubit2, np.pi / 2))
    circuit.append(tgqs.sqrt_x(qubit2))
    circuit.append(tgqs.rz(qubit2, np.pi / 2))

    circuit.append(tgqs.rz(qubit3, np.pi / 2))
    circuit.append(tgqs.sqrt_x(qubit3))
    circuit.append(tgqs.rz(qubit3, np.pi / 2))
    circuit.append(tgqs.cz(qubit2, qubit3))
    circuit.append(tgqs.rz(qubit3, np.pi / 2))
    circuit.append(tgqs.sqrt_x(qubit3))
    circuit.append(tgqs.rz(qubit3, np.pi / 2))

    circuit.append(tgqs.rz(qubit2, np.pi / 2))
    circuit.append(tgqs.sqrt_x(qubit2))
    circuit.append(tgqs.rz(qubit2, np.pi / 2))
    circuit.append(tgqs.cz(qubit3, qubit2))
    circuit.append(tgqs.rz(qubit2, np.pi / 2))
    circuit.append(tgqs.sqrt_x(qubit2))
    circuit.append(tgqs.rz(qubit2, np.pi / 2))

    circuit.append(tgqs.rz(qubit1, np.pi / 2))
    circuit.append(tgqs.sqrt_x(qubit1))
    circuit.append(tgqs.rz(qubit1, np.pi / 2))
    circuit.append(tgqs.cz(qubit2, qubit1))
    circuit.append(tgqs.rz(qubit1, np.pi / 2))
    circuit.append(tgqs.sqrt_x(qubit1))
    circuit.append(tgqs.rz(qubit1, np.pi / 2))
    circuit.append(tgqs.rz(qubit1, -np.pi/4))
    circuit.append(tgqs.rz(qubit2, np.pi/4))
    # circuit.append(tgqs.CNOT(qubit2, qubit1))
    circuit.append(tgqs.rz(qubit1, np.pi / 2))
    circuit.append(tgqs.sqrt_x(qubit1))
    circuit.append(tgqs.rz(qubit1, np.pi / 2))
    circuit.append(tgqs.cz(qubit2, qubit1))
    circuit.append(tgqs.rz(qubit1, np.pi / 2))
    circuit.append(tgqs.sqrt_x(qubit1))
    circuit.append(tgqs.rz(qubit1, np.pi / 2))
    return circuit
